@@ -1,23 +1,29 @@
+"""Lightweight FreeIPA JSON RPC client."""
+
 import json
 import logging
+
 import requests
 
-from .exceptions import (Unauthorized, DuplicateEntry, FreeIPAError)
-from .exceptions import (parse_group_management_error, parse_error)
+from .exceptions import (
+    DuplicateEntry, FreeIPAError, Unauthorized,
+    parse_error, parse_group_management_error,
+    parse_hostgroup_management_error
+)
 
 logger = logging.getLogger(__name__)
 
 
 class Client(object):
-    """
-    Lightweight FreeIPA JSON RPC client.
-    """
+    """Lightweight FreeIPA JSON RPC client."""
 
     def __init__(self, host, verify_ssl=True, version=None):
         """
+        Initialize client with connection options.
+
         :param host: hostname to connect to
         :type host: string
-        :param verify_ssl: verify SSL certificates for HTTPS requests, defaults to True
+        :param verify_ssl: verify SSL certificates for HTTPS requests
         :type verify_ssl: bool
         :param version: default client version, may be overwritten in individual requests
         :type version: string
@@ -36,7 +42,7 @@ class Client(object):
         :type username: string
         :param password: password of the user
         :type password: string
-        :raises Unauthorized:
+        :raises Unauthorized: raised if credentials are invalid.
         """
         login_url = '{0}/session/login_password'.format(self._base_url)
         headers = {
@@ -110,8 +116,36 @@ class Client(object):
         else:
             return result['result']
 
+    def change_password(self, username, new_password, old_password):
+        """
+        Set the password of a user. (Does not expire)
+
+        :param username: User login (username)
+        :type username: string
+        :param new_password: New password for the user
+        :type new_password: string
+        :param old_password: Users old password
+        :type old_password: string
+        """
+
+        password_url = '{0}/session/change_password'.format(self._base_url)
+        headers = {
+            'Content-Type': 'application/x-www-form-urlencoded',
+            'Accept': 'text/plain'
+        }
+
+        data = {'user': username, 'new_password': new_password, 'old_password': old_password}
+        response = self._session.post(password_url, headers=headers, data=data, verify=self._verify_ssl)
+
+        if not response.ok:
+            raise FreeIPAError(message=response.content, code=response.status_code)
+
+        if response.headers.get('X-IPA-Pwchange-Result', None) != 'ok':
+            raise FreeIPAError(message=response.content, code=response.status_code)
+        return response
+
     def user_add(self, username, first_name, last_name, full_name, display_name=None,
-                 noprivate=False, mail=None, ssh_key=None, job_title=None,
+                 noprivate=False, mail=None, ssh_key=None, job_title=None, gid_number=None, uid_number=None,
                  preferred_language=None, disabled=False, random_pass=False, initials=None, home_directory=None,
                  gecos=None, login_shell=None, user_password=None, street_address=None, city=None, state=None,
                  postal_code=None, telephone_number=None, mobile_number=None, pager_number=None, fax_number=None,
@@ -139,6 +173,10 @@ class Client(object):
         :type ssh_key: string or list
         :param job_title: Job title
         :type job_title: string
+        :param gid_number: gidNumber
+        :type gid_number: string
+        :param uid_number: uidNumber
+        :type uid_number: string
         :param preferred_language: Preferred language ISO code
         :type preferred_language: string
         :param disabled: Account disabled
@@ -149,7 +187,7 @@ class Client(object):
         :type initials: string
         :param home_directory: Home directory field of user
         :type home_directory: string
-        :param gecos: Gecos field of user
+        :param gecos: GECOS field is a comma-delimited list used to record general information about the user.
         :type gecos: string
         :param login_shell: Login shell field of user
         :type login_shell: string
@@ -190,7 +228,7 @@ class Client(object):
         :param employee_number: Employee number of user
         :type employee_number: string
         :param employee_type: Employee type of user
-        :type employee_type string
+        :type employee_type: string
         """
         params = {
             'all': True,
@@ -199,6 +237,12 @@ class Client(object):
             'cn': full_name,
         }
 
+        if gid_number:
+            params['gidnumber'] = gid_number
+
+        if uid_number:
+            params['uidnumber'] = uid_number
+
         if display_name:
             params['displayname'] = display_name
 
@@ -430,7 +474,7 @@ class Client(object):
         :param employee_number: Employee number of user
         :type employee_number: string
         :param employee_type: Employee type of user
-        :type employee_type string
+        :type employee_type: string
         """
         params = {
             'all': False,  # Retrieve and print all attributes from the server.
@@ -561,6 +605,15 @@ class Client(object):
         }
         return self._request('user_del', username, params)
 
+    def user_undel(self, username):
+        """
+        Undelete a user.
+
+        :param username: User login.
+        :type username: string
+        """
+        return self._request('user_undel', username)
+
     def passwd(self, login, password, current_password=None):
         """
         Set the password of a user.
@@ -587,7 +640,7 @@ class Client(object):
         Create a new group.
 
         :param group: Group name, it should be alphanumeric and maximum length is 255.
-        :tyoe group: string
+        :type group: string
         :param description: Group description
         :type description: string
         :param non_posix: Create as non-POSIX group
@@ -746,18 +799,19 @@ class Client(object):
 
     def automountkey_find(self, location, automount_map, key=None, criteria=None, **kwargs):
         """
-        Search for automount key
+        Search for the automount key.
 
-        :params key:
+        :param key: Automount key name.
         :type key: string
-        :params automount_location: Automount location name
-        :type automount_location: string
-        :params automount_map: Automount map name
+        :param location: Automount location name
+        :type location: string
+        :param automount_map: Automount map name
         :type automount_map: string
+        :param criteria: A string searched in all relevant object attributes.
+        :type criteria: string
         :return:
         :rtype: dict
         """
-
         args = [
             location,
             automount_map
@@ -779,23 +833,22 @@ class Client(object):
         data = self._request('automountkey_find', args, params)
         return data['result']
 
-    def automounkey_add(self, key, mount_info, location, automount_map, **kwargs):
+    def automountkey_add(self, key, mount_info, location, automount_map, **kwargs):
         """
-        Create a new automount key
+        Create a new automount key.
 
         :param key: Automount key name
         :type key: string
-        :params mount_info: Mount information
+        :param mount_info: Mount information
         :type mount_info: string
-        :params location: Automount location name
-        :type string
-        :params automount_map: Automount map name
-        :type string
-        :returns: automount key data
+        :param location: Automount location name
+        :type location: string
+        :param automount_map: Automount map name
+        :type automount_map: string
+        :return: automount key data
         :rtype: dict
         """
-
-        if self.automountkey_find(key, location, automount_map):
+        if self.automountkey_find(location, automount_map, key):
             raise DuplicateEntry()
 
         args = [
@@ -814,22 +867,21 @@ class Client(object):
         data = self._request('automountkey_add', args, params)
         return data
 
-    def automounkey_mod(self, key, mount_info, automount_location, automount_map):
+    def automountkey_mod(self, key, mount_info, automount_location, automount_map):
         """
         Modify an automount key.
 
         :param key: Automount key name
         :type key: string
-        :params mount_info: Mount information
+        :param mount_info: Mount information
         :type mount_info: string
-        :params automount_location: Automount location name
-        :type string
-        :params automount_map: Automount map name
-        :type string
-        :returns: automount key data
+        :param automount_location: Automount location name
+        :param: type string
+        :param automount_map: Automount map name
+        :type automount_map: string
+        :return: automount key data
         :rtype: dict
         """
-
         args = [
             automount_location,
             automount_map
@@ -903,17 +955,16 @@ class Client(object):
             'rights': False
         }
 
-        data = self._request('automountlocation_find', location, params)
+        data = self._request('automountlocation_show', location, params)
         return data['result']
 
     def automountlocation_tofiles(self, location):
         """
         Generate automount files for a specific location.
 
-        :params location: Automount location name
-        :type string
+        :param location: Automount location name
+        :type location: string
         """
-
         data = self._request('automountlocation_tofiles', location)
         return data['result']
 
@@ -921,10 +972,10 @@ class Client(object):
         """
         Display an automount map.
 
-        :params location: Automount location name
-        :type string
-        :params automount_map: Automount map name
-        :type string
+        :param location: Automount location name
+        :type location: string
+        :param automount_map: Automount map name
+        :type automount_map: string
         """
         args = [
             location,
@@ -944,10 +995,10 @@ class Client(object):
         """
         Delete an automount map.
 
-        :params location: Automount location name
-        :type string
-        :params automount_map: Automount map name
-        :type string
+        :param location: Automount location name
+        :type location: string
+        :param automount_map: Automount map name
+        :type automount_map: string
         """
         args = [location, automount_map]
         params = {'continue': skip_errors}
@@ -959,8 +1010,8 @@ class Client(object):
         """
         Find an automount map.
 
-        :params location: Automount location name
-        :type string
+        :param location: Automount location name
+        :type location: string
         :param criteria: A string searched in all relevant object attributes.
         :type criteria: string
         """
@@ -982,10 +1033,10 @@ class Client(object):
         """
         Modify an automount map.
 
-        :params location: Automount location name
-        :type string
-        :params automount_map: Automount map name
-        :type string
+        :param location: Automount location name
+        :type location: string
+        :param automount_map: Automount map name
+        :type automount_map: string
         """
         args = [
             location,
@@ -999,7 +1050,7 @@ class Client(object):
         }
 
         if description:
-            params.append(description)
+            params['description'] = description
 
         params.update(kwargs)
 
@@ -1010,10 +1061,10 @@ class Client(object):
         """
         Display an automount map.
 
-        :params location: Automount location name
-        :type string
-        :params automount_map: Automount map name
-        :type string
+        :param location: Automount location name
+        :type location: string
+        :param automount_map: Automount map name
+        :type automount_map: string
         """
         args = [
             location,
@@ -1029,3 +1080,412 @@ class Client(object):
 
         data = self._request('automountmap_show', args, params)
         return data['result']
+
+    def host_add(self, host, **kwargs):
+        """
+        Create a new host.
+
+        :param host: Host name which should be alphanumeric and maximum length is 255
+        :type host: string
+        """
+        params = {
+            'all': True
+        }
+        params.update(kwargs)
+
+        data = self._request('host_add', host, params)
+        return data['result']
+
+    def host_del(self, fqdn, skip_errors=False, updatedns=None):
+        """
+        Delete host from FreeIPA
+
+        :param fqdn: Host name
+        :type fqdn: string
+        :param skip_errors: Continuous mode: Don't stop on errors
+        :type skip_errors: bool
+        :param updatedns: Remove A, AAAA, SSHFP and PTR records of the host(s) managed by IPA DNS
+        :type updatedns: bool
+        """
+        params = {
+            "continue": skip_errors
+        }
+
+        if updatedns:
+            params['updatedns'] = updatedns
+
+        data = self._request('host_del', fqdn, params)
+        return data['result']
+
+    def host_find(self, criteria=None, allattr=True, no_members=False, sizelimit=0, raw=False, **kwargs):
+        """
+        Search for hosts.
+
+        :param criteria: A string searched in all relevant object attributes.
+        :type criteria: string
+        :param allattr: Retrieve and print all attributes from the server. Affects command output
+        :type allattr: bool
+        :param no_members: Suppress processing of membership attributes
+        :type no_members: bool
+        :param sizelimit: Maximum number of entries returned (0 is unlimited)
+        :type sizelimit: int
+        :param raw: Print entries as stored on the server. Only affects output format
+        :type raw: bool
+        """
+        params = {
+            'all': allattr,             # Retrieve and print all attributes from the server. Affects command output
+            'no_members': no_members,   # Suppress processing of membership attributes
+            'sizelimit': sizelimit,     # Maximum number of entries returned (0 is unlimited)
+            'raw': raw                  # Print entries as stored on the server. Only affects output format
+        }
+        params.update(kwargs)
+        return self._request('host_find', criteria, params)
+
+    def host_show(self, fqdn, rights=False, no_members=False, allattr=True, raw=False):
+        """
+        Display information about a host.
+
+        :param fqdn: Host name
+        :type fqdn: string
+        :param rights: Display the access rights of this entry (requires --all). See ipa man page for details
+        :type rights: bool
+        :param no_members: Suppress processing of membership attributes
+        :type no_members: bool
+        :param allattr: Retrieve and print all attributes from the server. Affects command output
+        :type allattr: bool
+        :param raw: Print entries as stored on the server. Only affects output format
+        :type raw: bool
+        """
+        data = self._request('host_show', fqdn, {'all': allattr, 'rights': rights, 'no_members': no_members,
+                                                 'raw': raw})
+        return data['result']
+
+    def hostgroup_add(self, hostgroup, description=None, no_members=False, **kwargs):
+        """
+        Create a new host group.
+
+        :param hostgroup: Host Group name, it should be alphanumeric and maximum length is 255.
+        :type hostgroup: string
+        :param description: Host Group description
+        :type description: string
+        :param no_members: Suppress processing of membership attributes
+        :type no_members: bool
+        """
+        params = {'all': True}
+
+        if description:
+            params['description'] = description
+
+        if no_members:
+            params['no_members'] = no_members
+
+        params.update(kwargs)
+        data = self._request('hostgroup_add', hostgroup, params)
+        return data['result']
+
+    def hostgroup_del(self, hostgroup_name, skip_errors=False):
+        """
+        Delete a hostgroup
+
+        :param hostgroup_name: Name of hostgroup
+        :type hostgroup_name: string
+        :param skip_errors: Continuous mode: Don't stop on errors
+        :type skip_errors: bool
+        """
+
+        data = self._request('hostgroup_del', hostgroup_name, {'continue': skip_errors})
+        return data['result']
+
+    def hostgroup_find(self, criteria=None, allattr=True, no_members=False, sizelimit=0, raw=False, **kwargs):
+        """
+        Search for hostgroups
+
+        :param criteria: A string searched in all relevant object attributes.
+        :type criteria: string
+        :param allattr: Retrieve and print all attributes from the server. Affects command output
+        :type allattr: bool
+        :param no_members: Suppress processing of membership attributes
+        :type no_members: bool
+        :param sizelimit: Maximum number of entries returned (0 is unlimited)
+        :type sizelimit: int
+        :param raw: Print entries as stored on the server. Only affects output format
+        :type raw: bool
+        """
+        params = {
+            'all': allattr,             # Retrieve and print all attributes from the server. Affects command output
+            'no_members': no_members,   # Suppress processing of membership attributes
+            'sizelimit': sizelimit,     # Maximum number of entries returned (0 is unlimited)
+            'raw': raw                  # Print entries as stored on the server. Only affects output format
+        }
+        params.update(kwargs)
+        return self._request('hostgroup_find', criteria, params)
+
+    def hostgroup_show(self, hostgroup, rights=False, no_members=False, allattr=True, raw=False):
+        """
+        Display information about a host.
+
+        :param hostgroup: Hostgroup name
+        :type hostgroup: string
+        :param rights: Display the access rights of this entry (requires --all). See ipa man page for details
+        :type rights: bool
+        :param no_members: Suppress processing of membership attributes
+        :type no_members: bool
+        :param allattr: Retrieve and print all attributes from the server. Affects command output
+        :type allattr: bool
+        :param raw: Print entries as stored on the server. Only affects output format
+        :type raw: bool
+        """
+        data = self._request('hostgroup_show', hostgroup, {'all': allattr, 'rights': rights, 'no_members': no_members,
+                             'raw': raw})
+        return data['result']
+
+    def hostgroup_mod(self, hostgroup, description=None, no_members=False, rights=False, allattr=False, raw=False,
+                      **kwargs):
+        """
+        Modify a hostgroup.
+
+        :param hostgroup: Group name.
+        :type hostgroup: string
+        :param description: Group description
+        :type description: string
+        :param no_members: Suppress processing of membership attributes
+        :type no_members: bool
+        :param rights: Display the access rights of this entry (requires --all). See ipa man page for details
+        :type rights: bool
+        :param allattr: Retrieve and print all attributes from the server. Affects command output
+        :type allattr: bool
+        :param raw: Print entries as stored on the server. Only affects output format
+        :type raw: bool
+        """
+        params = {
+            'all': allattr,
+            'raw': raw,
+            'rights': rights,
+            'no_members': no_members
+        }
+        if description:
+            params['description'] = description
+
+        params.update(kwargs)
+        data = self._request('hostgroup_mod', hostgroup, params)
+        return data['result']
+
+    def hostgroup_add_members(self, hostgroup, no_members=False, host=None, hostgroups=None, skip_errors=False,
+                              **kwargs):
+        """
+        Add members to a hostgroup.
+
+        :param hostgroup: Host Group name, it should be alphanumeric and maximum length is 255.
+        :type hostgroup: string
+        :param no_members: Suppress processing of membership attributes
+        :type no_members: bool
+        :param host: Hosts to add
+        :type host: list or string
+        :param hostgroups: Host group to add
+        :type hostgroups: list or string
+        :param skip_errors: Skip processing errors.
+        :type skip_errors: bool
+        """
+        params = {'all': True}
+
+        if no_members:
+            params['no_members'] = no_members
+
+        if host:
+            params['host'] = host
+
+        if hostgroups:
+            params['hostgroup'] = hostgroups
+
+        params.update(kwargs)
+        data = self._request('hostgroup_add_member', hostgroup, params)
+        if not skip_errors:
+            parse_hostgroup_management_error(data)
+        return data['result']
+
+    def hostgroup_remove_members(self, hostgroup, no_members=False, host=None, hostgroups=None, skip_errors=False,
+                                 **kwargs):
+        """
+        Remove members from a hostgroup.
+
+        :param hostgroup: Host Group name, it should be alphanumeric and maximum length is 255.
+        :type hostgroup: string
+        :param no_members: Suppress processing of membership attributes
+        :type no_members: bool
+        :param host: Hosts to remove
+        :type host: list or string
+        :param hostgroups: Host group to remove
+        :type hostgroups: list or string
+        :param skip_errors: Skip processing errors.
+        :type skip_errors: bool
+        """
+        params = {'all': True}
+
+        if no_members:
+            params['no_members'] = no_members
+
+        if host:
+            params['host'] = host
+
+        if hostgroups:
+            params['hostgroup'] = hostgroups
+
+        params.update(kwargs)
+        data = self._request('hostgroup_remove_member', hostgroup, params)
+        if not skip_errors:
+            parse_hostgroup_management_error(data)
+        return data['result']
+
+    def dnsrecord_add(self, zone_name, record_name, **kwargs):
+        """
+        Create a new DNS record.
+
+        :param zone_name: DNS zone name (e.g., example.com)
+        :type zone_name: string
+        :param record_name: DNS record name (e.g., host1)
+        :type record_name: string
+        """
+        params = {'all': True}
+
+        params.update(kwargs)
+        data = self._request('dnsrecord_add', [zone_name, record_name], params)
+        return data['result']
+
+    def dnsrecord_del(self, zone_name, record_name, **kwargs):
+        """
+        Delete a DNS record.
+
+        :param zone_name: DNS zone name (e.g., example.com)
+        :type zone_name: string
+        :param record_name: DNS record name (e.g., host1)
+        :type record_name: string
+        """
+        params = {}
+
+        params.update(kwargs)
+        self._request('dnsrecord_del', [zone_name, record_name], params)
+
+    def dnsrecord_find(self, zone_name, criteria=None, **kwargs):
+        """
+        Search for DNS records.
+
+        :param zone_name: DNS zone name (e.g., example.com)
+        :type zone_name: string
+        :param criteria: A string searched in all relevant object attributes.
+        :type criteria: string
+        """
+        params = {
+            'all': True,
+            'sizelimit': 0
+        }
+        params.update(kwargs)
+        return self._request('dnsrecord_find', [zone_name, criteria], params)
+
+    def dnsrecord_show(self, zone_name, record_name, **kwargs):
+        """
+        Display information about a DNS record.
+
+        :param zone_name: DNS zone name (e.g., example.com)
+        :type zone_name: string
+        :param record_name: DNS record name (e.g., host1)
+        :type record_name: string
+        """
+        params = {
+            'all': True,
+            'raw': False,
+            'rights': False,
+        }
+        params.update(kwargs)
+        data = self._request('dnsrecord_show', [zone_name, record_name], params)
+        return data['result']
+
+    def dnsrecord_mod(self, zone_name, record_name, **kwargs):
+        """
+        Modify a DNS record.
+
+        :param zone_name: DNS zone name (e.g., example.com)
+        :type zone_name: string
+        :param record_name: DNS record name (e.g., host1)
+        :type record_name: string
+        """
+        params = {
+            'all': False,
+            'raw': False,
+            'rights': False,
+        }
+
+        params.update(kwargs)
+        data = self._request('dnsrecord_mod', [zone_name, record_name], params)
+        return data['result']
+
+    def dnszone_add(self, zone_name, **kwargs):
+        """
+        Create a new DNS zone.
+
+        :param zone_name: DNS zone name (e.g., example.com)
+        :type zone_name: string
+        """
+        params = {'all': True}
+
+        params.update(kwargs)
+        data = self._request('dnszone_add', zone_name, params)
+        return data['result']
+
+    def dnszone_del(self, zone_name, **kwargs):
+        """
+        Delete a DNS zone.
+
+        :param zone_name: DNS zone name (e.g., example.com)
+        :type zone_name: string
+        """
+        params = {}
+
+        params.update(kwargs)
+        self._request('dnszone_del', zone_name, params)
+
+    def dnszone_find(self, criteria=None, **kwargs):
+        """
+        Search for DNS zones.
+
+        :param criteria: A string searched in all relevant object attributes.
+        :type criteria: string
+        """
+        params = {
+            'all': True,
+            'sizelimit': 0
+        }
+        params.update(kwargs)
+        return self._request('dnszone_find', criteria, params)
+
+    def dnszone_show(self, zone_name, **kwargs):
+        """
+        Display information about a DNS zone.
+
+        :param zone_name: DNS zone name (e.g., example.com)
+        :type zone_name: string
+        """
+        params = {
+            'all': True,
+            'raw': False,
+            'rights': False,
+        }
+        params.update(kwargs)
+        data = self._request('dnszone_show', zone_name, params)
+        return data['result']
+
+    def dnszone_mod(self, zone_name, **kwargs):
+        """
+        Modify a DNS zone.
+
+        :param zone_name: DNS zone name (e.g., example.com)
+        :type zone_name: string
+        """
+        params = {
+            'all': False,
+            'raw': False,
+            'rights': False,
+        }
+
+        params.update(kwargs)
+        data = self._request('dnszone_mod', zone_name, params)
+        return data['result']
